
function [xopt, iter, time, fv, nD, sparsity, avar, fs] = Driver_AManPG(Xinitial, A, Dsq, lambda, Ll, Lu, tol, maxiter, Ftol)
    % Min_{X \in Ob} \| X^T A^T A X - D^2\|_F^2 + lambda \|X\|_1
    % functions for the optimization problem
    fhandle = @(x)f(x, A, Dsq, lambda);
    gfhandle = @(x)gf(x, A, Dsq, lambda);
    fprox = @prox;
    fcalJ = @calJ;
    
    % functions for the manifold
    fcalA = @calA;
    fcalAstar = @calAstar;
    
    [xopt, iter, time, fv, nD, fs] = solver(fhandle, gfhandle, fcalA, fcalAstar, fprox, fcalJ, Xinitial, Ll, Lu, lambda, tol, maxiter, Ftol);
    sparsity = sum(sum(abs(xopt.main) < 1e-5)) / prod(size(xopt.main));
%     fprintf('sparsity:%1.3f\n',sum(sum(abs(xopt.main) < 1e-5)) / (n * p));

    % adjusted variance
    [Q, R] = qr(A * xopt.main, 0);
    avar = trace(R * R);
    
end

function output = VT(x, y, xix)
    output = zeros(size(x.main));
    for i = 1 : size(x.main, 2)
        output(:, i) = xix(:, i) - (2 * (xix(:, i)' * y.main(:, i)) / norm(x.main(:, i) + y.main(:, i))^2) * (x.main(:, i) + y.main(:, i));
    end
end

function output = Rinv(x, y)
    output = zeros(size(x.main));
    for i = 1 : size(output, 2)
        tmp = x.main(:, i)' * y.main(:, i);
        if(abs(tmp - 1) < eps)
            output(:, i) = (y.main(:, i) - x.main(:, i) * tmp);
        else
            output(:, i) = (y.main(:, i) - x.main(:, i) * tmp) * (acos(tmp) / sqrt(1 - tmp * tmp));
        end
    end
end

% use exponential map
function output = R(x, eta)
    output.main = zeros(size(eta));
    for i = 1 : size(output.main, 2)
        tmp = norm(eta(:, i));
        if(abs(tmp) < eps)
            output.main(:, i) = x.main(:, i) * cos(tmp) + eta(:, i);
        else
            output.main(:, i) = x.main(:, i) * cos(tmp) + eta(:, i) / tmp * sin(tmp);
        end
        output.main(:, i) = output.main(:, i) / norm(output.main(:, i));
    end
end


% function output = R(x, eta)
%     output.main = x.main + eta;
%     for i = 1 : size(output.main, 2)
%         output.main(:, i) = output.main(:, i) / norm(output.main(:, i));
%     end
% end

function [output, x] = f(x, A, Dsq, lambda)
    X = x.main;
    x.AX = A * X;
    x.XtAtAXmDsq = (x.AX' * x.AX - diag(Dsq));
    output = norm(X(:), 1) * lambda + norm(x.XtAtAXmDsq , 'fro')^2;
end

function output = gf(x, A, Dsq, lambda)
    X = x.main;
    Egf = 4 * A' * (x.AX * x.XtAtAXmDsq);
    output = Egf - X * diag(diag(X' * Egf));
end

function output = prox(X, t, mu)
    output = min(0, X + t * mu) + max(0, X - t * mu);
end

function output = calA(Z, U) % U \in St(p, n)
    tmp = Z' * U;
    output = tmp + tmp';
end

function output = calAstar(Lambda, U) % U \in St(p, n)
    output = U * (Lambda + Lambda');
end

function output = calJ(y, eta, t, mu)
    output = (abs(y) > mu * t) .* eta;
end

function [xopt, iter, time, fv, err, fs] = solver(fhandle, gfhandle, fcalA, fcalAstar, fprox, fcalJ, x0, Ll, Lu, mu, tol, maxiter, Ftol)
    delta = 0.0001;
    gamma = 0.5;
    err = inf;
    x1 = x0;
    y1 = x0;
    [fx1, x1] = fhandle(x1);
    gfx1 = gfhandle(x1);
    t = 1 / Ll;
    t0 = 1 / Lu;
    iter = 0;
    fs(iter + 1) = fx1;
    [n, p] = size(x0.main);
    s1 = 1;
    Dyinitial = zeros(p, p);
    Dxinitial = zeros(p, p);
    totalbt = 0;
    innertol = max(1e-13, min(1e-11,1e-3*sqrt(tol)/ (Lu^2)));
    gfy1 = gfx1;
    fx2 = fx1;
    fx0 = fx1;
    gfx0 = gfx1;
    fprintf('iter:%d, f:%e, ngf:%e\n', iter, fx1, norm(gfx1, 'fro'));
    N = 5;
    maxN = 10;
    minN = 2;
    justrestart = 0;
    SafeGuardIter = N;
    Dx = inf;
    btiter = 0;
    tic;
    while(err > tol && fx1 > Ftol + 1e-7 && iter < maxiter)
        % safeguard
        if(iter == SafeGuardIter)
            if(justrestart == 1)
                Dx = Dybackup;
                Dxinitial = Dxinitialbackup;
                justrestart = 0;
            else
                [Dx, Dxinitial, inneriterx] = finddir(x0, gfx0, t, fcalA, fcalAstar, fprox, fcalJ, mu, Dxinitial, innertol);
            end
            alpha = 1;
            xc = R(x0, alpha * Dx);
            [fxc, xc] = fhandle(xc);
            btiter = 0;
            while(fxc > fx0 - delta * alpha * norm(Dx, 'fro')^2 && btiter < 3)
                alpha = alpha * gamma;
                xc = R(x0, alpha * Dx);
                [fxc, xc] = fhandle(xc);
                btiter = btiter + 1;
                totalbt = totalbt + 1;
            end
            if(btiter == 3)
                innertol = max(innertol * 1e-2, 1e-20);
                t = t / 1.1;
                continue;
            end
            % if safeguard takes effect, here must be a strict inequality,
            % otherwise, algorithm would stack at N = 1.
            if(iter ~= 0 && fxc < fx2)
                gfxc = gfhandle(xc);
                y1 = xc;gfy1 = gfxc; fy1 = fxc;
                x1 = xc; fx1 = fxc; gfx1 = gfxc;
                s1 = 1;
                if(N ~= maxN)
                    t = max(t0, t / 1.1);
                end
                N = max(N - 1, minN);
                justrestart = 1;
            else
                N = min(N + 1, maxN);
            end
            % update x0
            x0 = x1;
            fx0 = fx1;
            gfx0 = gfhandle(x0);
        end
        [Dy, Dyinitial, inneritery] = finddir(y1, gfy1, t, fcalA, fcalAstar, fprox, fcalJ, mu, Dyinitial, innertol);
        if(justrestart == 1 && iter == SafeGuardIter) % this is to avoid a duplicate computation in the safeguard
            Dybackup = Dy;
            Dxinitialbackup = Dyinitial;
        end
        if(iter == SafeGuardIter) % find the next safeguard iteration.
            SafeGuardIter = SafeGuardIter + N;
        end
        alpha = 1;
        x2 = R(y1, alpha * Dy);
        [fx2, x2] = fhandle(x2);
        
        s2 = (1 + sqrt(1 + 4 * s1 * s1)) / 2;
        zeta = (s2 + s1 - 1) / s2 * Dy - (s1 - 1) / s2 * Rinv(y1, x1);
%         zeta = (- (s1 - 1) / s2) * Rinv(x2, x1);

        iter = iter + 1;
        fs(iter + 1) = min(fx2, fhandle(x0));
        err = (norm(Dx, 'fro') / t0)^2;
        Ds(iter) = norm(Dx, 'fro');
%         fprintf('iter:%d, f:%e, nD^2:%e, inneriterx:%d, inneritery:%d, btiter:%d\n', iter, fx2, err, inneriterx, inneritery, btiter);
        
%         y2 = R(x2, zeta);
        y2 = R(y1, zeta);
        [fy2, y2] = fhandle(y2);
        gfy2 = gfhandle(y2);
        fy1 = fy2; gfy1 = gfy2;
        y1 = y2;
        s1 = s2;
        x1 = x2; fx1 = fx2;
        if(mod(iter, 500) == 0)
%             [fx2, x2] = fhandle(x2);
            fprintf('iter:%d, fx:%e, N:%d, btiter:%d, err:%e\n', iter, min(fx2, fhandle(x0)), N, btiter, err);
        end
    end
    fprintf('iter:%d, fx:%e, err:%e, totalbt:%d\n', iter, min(fx2, fhandle(x0)), err, totalbt);
    xopt = x0;
    fv = fx0;
    time = toc;
end

% compute E(Lambda)
function ELambda = E(Lambda, BLambda, x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu)
    if(length(BLambda) == 0)
        BLambda = x - t * (gfx - fcalAstar(Lambda, x));
    end
    DLambda = fprox(BLambda, t, mmu) - x;
    ELambda = fcalA(DLambda, x);
end

% compute calG(Lambda)[d]
function GLambdad = GLd(Lambda, d, BLambda, Blocks, x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu)
        GLambdad = t * fcalA(fcalJ(BLambda, fcalAstar(d, x), t, mmu), x);
end

function [output, Lambda, inneriter] = finddir(x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu, x0, innertol)
    output = zeros(size(x.main));
    Lambda = zeros(size(x0));
    inneriterall = zeros(size(x.main, 2), 1);
    for i = 1 : size(x.main, 2)
        [output(:, i), Lambda(i, i), inneriterall(i)] = finddirSt(x.main(:, i), gfx(:, i), t, fcalA, fcalAstar, fprox, fcalJ, mmu, x0(i, i), innertol);
    end
    inneriter = mean(inneriterall);
end

% Use semi-Newton to solve the subproblem and find the search direction
function [output, Lambda, inneriter] = finddirSt(x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu, x0, innertol)
    lambda = 0.2;
    nu = 0.99;
    tau = 0.1;
    eta1 = 0.2; eta2 = 0.75;
    gamma1 = 3; gamma2 = 5;
    alpha = 0.1;
    beta = 1 / alpha / 100;
    [n, p] = size(x);
    
    z = x0;
    BLambda = x - t * (gfx - fcalAstar(z, x));
    Fz = E(z, BLambda, x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu);
    
    nFz = norm(Fz, 'fro');
    nnls = 5;
    xi = zeros(nnls, 1);% for non-monotonic linesearch
    xi(nnls) = nFz;
    maxiter = 1000;
    times = 0;
    Blocks = cell(p, 1);
    while(nFz * nFz > innertol && times < maxiter) % while not converge, find d and update z
        mu = lambda * max(min(nFz, 0.1), 1e-11);
        Axhandle = @(d)GLd(z, d, BLambda, Blocks, x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu) + mu * d;
        [d, CGiter] = myCG(Axhandle, -Fz, tau, lambda * nFz, 30); % update d
        u = z + d;
        Fu = E(u, [], x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu); 
        nFu = norm(Fu, 'fro');
        
        if(nFu < nu * max(xi))
            z = u;
            Fz = Fu;
            nFz = nFu;
            xi(mod(times, nnls) + 1) = nFz;
            status = 'success';
        else
            rho = - sum(Fu(:) .* d(:)) / norm(d, 'fro')^2;
            if(rho >= eta1)
                v = z - sum(sum(Fu .* (z - u))) / nFu^2 * Fu;
                Fv = E(v, [], x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu);
                nFv = norm(Fv, 'fro');
                if(nFv <= nFz)
                    z = v;
                    Fz = Fv;
                    nFz = nFv;
                    status = 'safegard success projection';
                else
                    z = z - beta * Fz;
                    Fz = E(z, [], x, gfx, t, fcalA, fcalAstar, fprox, fcalJ, mmu);
                    nFz = norm(Fz, 'fro');
                    status = 'safegard success fixed-point';
                end
            else
%                 fprintf('unsuccessful step\n');
                status = 'safegard unsuccess';
            end
            if(rho >= eta2)
                lambda = max(lambda / 4, 1e-5);
            elseif(rho >= eta1)
                lambda = (1 + gamma1) / 2 * lambda;
            else
                lambda = (gamma1 + gamma2) / 2 * lambda;
            end
        end
        BLambda = x - t * (gfx - fcalAstar(z, x));
%         fprintf(['iter:%d, nFz:%f, xi:%f, ' status '\n'], times, nFz, max(xi));
        times = times + 1;
    end
    Lambda = z;
    inneriter = times;
    output = fprox(BLambda, t, mmu) - x;
end

function [output, k] = myCG(Axhandle, b, tau, lambdanFz, maxiter)
    x = zeros(size(b));
    r = b;
    p = r;
    k = 0;
    while(norm(r, 'fro') > tau * min(lambdanFz * norm(x, 'fro'), 1) && k < maxiter)
        Ap = Axhandle(p);
        alpha = r(:)' * r(:) / (p(:)' * Ap(:));
        x = x + alpha * p;
        rr0 = r(:)' * r(:);
        r = r - alpha * Ap;
        beta = r(:)' * r(:) / rr0;
        p = r + beta * p;
        k = k + 1;
    end
    output = x;
end
