
function [xoptmain, iter, time, fv, nD, sparsity, avar, fs, totalinneriter] = Driver_MPGWH(Xinitial, A, lambda, L, tol, maxiter, Ftol, RPMWHmaxiter, RPMWHtol)
% Min_{X \in St} -Tr(X^T A^T A X) + lambda \|X\|_1

    fhandle = @(x)f(x, A, lambda);
    gfhandle = @(x)gf(x, A, lambda);
    
    [xopt, iter, time, fv, nD, fs, totalinneriter] = MPGWH_solver(fhandle, gfhandle, Xinitial, lambda, L, tol, maxiter, Ftol, RPMWHmaxiter, RPMWHtol);
    
    xopt.main(abs(xopt.main) < 1e-5) = 0;
    sparsity = sum(sum(abs(xopt.main) < 1e-5)) / (prod(size(xopt.main)));
    xoptmain = xopt.main;
    % adjusted variance
    [Q, R] = qr(A * xopt.main, 0);
    avar = trace(R * R);
end

function [output, iter, time, fv, err, fs, totalinneriter] = MPGWH_solver(fhandle, gfhandle, x0, lambda, L, tol, maxiter, Ftol, RPMWHmaxiter, RPMWHtol)
    x1 = x0;
    iter = 0;
    err = inf;
    maxbtiter = 10;
    L0 = L;
    [fv1, x1] = fhandle(x1);
    Rgf1 = gfhandle(x1);
    fs(iter + 1) = fv1;
    fprintf('iter:%d, f:%e\n', iter, fv1);
    totalbt = 0;
    totalinneriter = 0;
    innertol = max(1e-13, min(1e-11,1e-3*sqrt(tol) / L / L));
%     innertol = 1e-20;
%     maxiter = 2;%%----
% maxiter = 70;
    tic;
    while(iter < maxiter && err > tol && fv1 > Ftol + 1e-7)
        [etax1, inneriter] = RPMWH(x1, Rgf1, L, lambda, RPMWHtol, false, RPMWHmaxiter, innertol);
        totalinneriter = totalinneriter + inneriter;
        
%         etax1.main' * x1.main
        
        x2 = etax1.y;
%         x2 = R(x1, etax1);
%         error('sf');%%----
% x2.main' * x2.main

        [fv2, x2] = fhandle(x2);
        
        alpha = 1;
        btiter = 0;
        netax1sq = innerprod(x1, etax1, etax1);
        while(fv2 - fv1 > - 0.0001 *  alpha * netax1sq && btiter < maxbtiter)
            alpha = alpha * 0.5;
            etax2.main = alpha * etax1.main;
            x2 = RExpStief(x1, etax2);
            [fv2, x2] = fhandle(x2);
            btiter = btiter + 1;
            totalbt = totalbt + 1;
        end
        if(btiter == maxbtiter)
            L = min(L0,L * 1.2);
            innertol = max(innertol * 1e-2, 1e-20);
            iter = iter + 1;
            fs(iter + 1) = fv1;
            continue;
%             fprintf('warning: line search reaches max number of iterations!\n');
        end
        Rgf2 = gfhandle(x2);
        
        iter = iter + 1;
        fs(iter + 1) = fv2;
        err = netax1sq * L * L;% norm((x2.main - x1.main) * L, 'fro')^2;
        if(mod(iter, 100) == 0)
            fprintf('iter:%d, f:%e, err:%e, L:%e, btiter:%d, totalinneriter:%d\n', iter, fv2, err, L, btiter, totalinneriter);
        end
        x1 = x2;
        Rgf1 = Rgf2;
        fv1 = fv2;
        if(btiter == 0)
            L = L/1.01;
        else
            L = min(L0,L * 1.01);
        end
    end
    fprintf('iter:%d, f:%e, err:%e, L:%e, totalbt:%d, totalinneriter:%d\n', iter, fv2, err, L, totalbt, totalinneriter);
    output = x1;
    fv = fv1;
    time = toc;
end

function output = innerprod(x, etax, xix)
    tmp = xix.main - 0.5 * x.main * (x.main' * xix.main);
    output = etax.main(:)' * tmp(:);
end

function [output, x, eta] = RExpStief(x, eta)
%% Quotient geometry
    [n, p] = size(x.main);
    [Q, R] = qr(eta.main - x.main * (x.main' * eta.main), 0);
    eta.Q = Q;
    eta.R = R;
    eta.Omega = x.main' * eta.main;
    tmp = expm([eta.Omega, -R'; R, zeros(p, p)]);
    tmp = tmp(:, 1 : p);
    output.main = [x.main, Q] * tmp;
end

function [output, x] = f(x, A, lambda)
    x.Ax = A * x.main;
    tmp = norm(x.Ax, 'fro');
    output = - tmp * tmp + lambda * sum(abs(x.main(:)));
end

function output = gf(x, A, lambda)
    if(~isfield(x, 'Ax'))
        x.Ax = A * x.main;
    end
    gfx = -2 * (A' * x.Ax);
    output.main = gfx - x.main * (gfx' * x.main);
%     tmp = gfx' * x.main;
%     output.main = gfx - x.main * ((tmp + tmp') / 2);
end
